def test_import_deeptile():

    """ Test that DeepTile can be imported."""

    import deeptile


def test_import_extensions():

    """ Test that extensions can be imported. """

    from deeptile import extensions
